#!/bin/sh -xe

./test1 "$srcdir/test.fstrm" 7
./test1 "$srcdir/test.fstrm" 15
./test1 "$srcdir/test.fstrm" 24
./test1 "$srcdir/test.fstrm" 44
./test1 "$srcdir/test.fstrm" 79
./test1 "$srcdir/test.fstrm" 134
